#!/usr/bin/env python

'''Hipshot converts a video file or series of photographs into
a single image simulating a long-exposure photograph.
'''

import cv
from numpy import copy

from avena import image, np, video


__author__ = 'Mansour Moufid'
__copyright__ = 'Copyright 2013-2015, Mansour Moufid'
__license__ = 'ISC'
__version__ = '0.7'
__email__ = 'mansourmoufid@gmail.com'
__status__ = 'Development'


_EXT = '.png'


def merge(frames, alpha, display=None):
    '''Average a list of frames with a weight of alpha,
    optionally display the process in an OpenCV NamedWindow.
    '''
    first = next(frames)
    acc = first * alpha
    for frame in frames:
        acc += frame * alpha
        if display:
            display_acc = copy(acc)
            display_acc = video._array_to_cv(display_acc)
            cv.ShowImage(display, display_acc)
            k = cv.WaitKey(1)
            k = k & 255
            if k == ord('q'):
                break
            elif k == ord('z'):
                acc.fill(0)
    return acc


if __name__ == '__main__':
    pass



#!/usr/bin/env python

from numpy import around, empty as _empty, mean, std
from numpy import int8, int16, int32, int64
from numpy import uint8, uint16, uint32, uint64
from numpy import float32, float64
from sys import float_info as _float_info


_eps = 10.0 * _float_info.epsilon

# Map of NumPy array type strings to types
_np_dtypes = {
    'int8':     int8,
    'int16':    int16,
    'int32':    int32,
    'int64':    int64,
    'uint8':    uint8,
    'uint16':   uint16,
    'uint32':   uint32,
    'uint64':   uint64,
    'float32':  float32,
    'float64':  float64,
}


_dtype_bounds = {
    'float32':  (0.0, 1.0),
    'float64':  (0.0, 1.0),
    'uint8':    (0, 255),
}


def from_uint8(array, dtype):
    new_array = array.astype(dtype)
    return new_array


def to_uint8(array):
    uint8_array = _empty(array.shape, dtype=uint8)
    around(array * 255, out=uint8_array)
    return uint8_array


def clip(array, bounds):
    '''Clip the values of an array to the given interval.'''
    (min, max) = bounds
    x = array < min + _eps
    y = array > max - _eps
    array[x] = min
    array[y] = max
    return


def normalize(array):
    '''Normalize an array to the interval [0,1].'''
    mu = mean(array)
    rho2 = std(array)
    min = mu - 3.0 * rho2
    max = mu + 3.0 * rho2
    array -= min
    if max - min > _eps:
        array /= max - min
    return


if __name__ == '__main__':
    pass



#!/usr/bin/env python

'''Read and write image and video files with OpenCV'''


from cv2 import cv
from numpy import asarray
from os.path import exists as _exists

from . import np
from . import utils


# Map of OpenCV image depths to NumPy array types
_cv_depths = {
    cv.IPL_DEPTH_8U:    'uint8',
    cv.IPL_DEPTH_8S:    'int8',
    cv.IPL_DEPTH_16U:   'uint16',
    cv.IPL_DEPTH_16S:   'int16',
    cv.IPL_DEPTH_32S:   'int32',
    cv.IPL_DEPTH_32F:   'float32',
    cv.IPL_DEPTH_64F:   'float64',
}


# Inverse mapping of the above
_cv_depths_inv = utils._invert_dict(_cv_depths)


# Map of colours to array indices
_OCV_RGB = {
    'B': 0,
    'G': 1,
    'R': 2,
}


def _template_image(image, depth):
    '''Return an empty image object of the same dimensions,
    colour depth, and number of channels as the given image.
    '''
    size = (image.width, image.height)
    channels = image.nChannels
    dup = cv.CreateImage(size, depth, channels)
    return dup


def _save_image(image, filename, random=False, ext=None):
    '''Save an image object and return the file name.

    If the parameter 'random' is True, use a random file name.
    '''
    if random:
        newfile = utils.rand_filename(filename, ext=ext)
    else:
        newfile = filename
    newimage = _template_image(image, cv.IPL_DEPTH_8U)
    cv.ConvertScaleAbs(image, newimage, scale=255)
    cv.SaveImage(newfile, newimage)
    return newfile


def num_frames(video_file):
    '''Return the number of frames in a video file.
    '''
    cap = cv.CaptureFromFile(video_file)
    if not cap:
        raise IOError('CaptureFromFile')
    n = cv.GetCaptureProperty(cap, cv.CV_CAP_PROP_FRAME_COUNT)
    return int(n)


def _cv_to_array(im, dtype=None):
    '''Return an OpenCV image object as a NumPy array.

    If a dtype is specified, return an array of that type.
    '''
    arr_data = im[:, :]
    arr_shape = (im.height, im.width, im.nChannels)
    if dtype is None:
        arr_dtype = np._np_dtypes[_cv_depths[im.depth]]
    else:
        arr_dtype = np._np_dtypes[dtype]
    arr = asarray(arr_data, dtype=arr_dtype)
    arr.shape = arr_shape
    return arr


def _array_to_cv(arr):
    '''Return a NumPy array as an OpenCV image object.
    '''
    utils.swap_rgb(arr, utils._PREFERRED_RGB, to=_OCV_RGB)
    im_channels = utils.depth(arr)
    swap = lambda x, y, z=1: (y, x)
    im_shape = swap(*arr.shape)
    im_size = arr.dtype.itemsize * im_channels * im_shape[0]
    im_depth = _cv_depths_inv[str(arr.dtype)]
    im = cv.CreateImageHeader(im_shape, im_depth, im_channels)
    cv.SetData(im, arr.tostring(), im_size)
    return im


DEFAULT_FRAME_ARRAY_DTYPE = 'float32'


_CV_CAM_PROPERTIES = {
    'width':    cv.CV_CAP_PROP_FRAME_WIDTH,
    'height':   cv.CV_CAP_PROP_FRAME_HEIGHT,
    'fps':      cv.CV_CAP_PROP_FPS,
}

DEFAULT_CAM_PROPERTY_RES = (640, 480)
DEFAULT_CAM_PROPERTY_FPS = 15


def _set_cam_properties(cap, props):
    for p in props:
        cv.SetCaptureProperty(
            cap,
            _CV_CAM_PROPERTIES[p],
            props[p]
        )


def get_frames(video_file=None,
               cam=-1, cam_res=None, cam_fps=None,
               as_array=True, dtype=None):
    '''Return a list of individual frames in a video file or
    webcam stream.

    If the parameter 'video_file' is None, frames are read
    from the default webcam.

    The optional parameter 'cam_res' specifies the webcam
    resolution as a (width, height) tuple in pixels, the
    default is (640, 480).

    The optional parameter 'cam_fps' specifies the webcam
    framerate in frames per second, the default is 15.

    If the parameter 'as_array' is True, return NumPy arrays.
    If dtype is specified, return NumPy arrays of that type.
    '''

    if dtype is None:
        dtype = DEFAULT_FRAME_ARRAY_DTYPE

    if cam_res is None:
        res = DEFAULT_CAM_PROPERTY_RES
    else:
        res = cam_res

    if cam_fps is None:
        fps = DEFAULT_CAM_PROPERTY_FPS
    else:
        fps = cam_fps

    if video_file:
        if not _exists(video_file):
            raise IOError(video_file)
        cap = cv.CaptureFromFile(video_file)
        if not cap:
            raise IOError('CaptureFromFile')
    else:
        cap = cv.CaptureFromCAM(cam)
        if not cap:
            raise IOError('CaptureFromCAM')
        cam_properties = {
            'width':    res[0],
            'height':   res[1],
            'fps':      fps,
        }
        _set_cam_properties(cap, cam_properties)

    while True:
        img = cv.QueryFrame(cap)
        if not img:
            break
        if as_array:
            img = _cv_to_array(img, dtype=dtype)
            np.normalize(img)
            utils.swap_rgb(img, _OCV_RGB, to=utils._PREFERRED_RGB)
        yield img


def _get_video_properties(filename):
    cap = cv.CaptureFromFile(filename)
    if not cap:
        raise IOError('CaptureFromFile')
    fps = cv.GetCaptureProperty(cap, cv.CV_CAP_PROP_FPS)
    width = cv.GetCaptureProperty(cap, cv.CV_CAP_PROP_FRAME_WIDTH)
    height = cv.GetCaptureProperty(cap, cv.CV_CAP_PROP_FRAME_HEIGHT)
    fps = float(fps)
    fps = min(fps, 30.0)
    fps = max(fps, 15.0)
    width = int(width)
    height = int(height)
    return (fps, width, height)


if __name__ == '__main__':
    pass



#!/usr/bin/env python

'''2D spatial filtering with the FFT'''


from functools import partial
from numpy import (
    array as _array,
    empty as _empty,
    multiply as _multiply,
    real as _real,
    vectorize as _vectorize,
)
from numpy.fft import (
    fftshift as _fftshift,
    ifftshift as _ifftshift,
    irfft2 as _irfft2,
    rfft2 as _rfft2,
)

from . import image, utils


def __in_circle(a, b, r, coords):
    x, y = coords
    d = (x - a) ** 2 + (y - b) ** 2
    if d < r ** 2:
        return 1.0
    else:
        return 0.0


def _indices_array(shape):
    m, n = shape
    indices = _array(
        [[(i, j) for j in range(n)] for i in range(m)],
        dtype=('f4,f4'),
    )
    return indices


def _low_pass_filter(shape, radius):
    m, n = shape
    indices = _indices_array(shape)
    _in_circle = partial(__in_circle, m // 2, n // 2, radius)
    _v_in_circle = _vectorize(_in_circle)
    return _v_in_circle(indices)


def _high_pass_filter(shape, radius):
    return 1.0 - _low_pass_filter(shape, radius)


def _filter(array, filter):
    z = _empty(array.shape, dtype=array.dtype)
    for i, c in enumerate(image.get_channels(array)):
        C = _rfft2(c)
        C = _fftshift(C)
        _multiply(C, filter, out=C)
        C = _ifftshift(C)
        c = _irfft2(C, s=c.shape)
        c = _real(c)
        if utils.depth(array) > 1:
            z[:, :, i] = c
        else:
            z = c
    return z


def _rshape(shape):
    m, n = shape
    if n % 2 == 0:
        return (m, (n // 2) + 1)
    else:
        return (m, (n + 1) // 2)


def lowpass(array, radius):
    '''Apply a 2D low-pass filter to an image array.'''
    rfilter = _low_pass_filter(_rshape(array.shape[:2]), radius)
    return _filter(array, rfilter)


def highpass(array, radius):
    '''Apply a 2D high-pass filter to an image array.'''
    m, n = array.shape[:2]
    rfilter = _high_pass_filter(_rshape(array.shape[:2]), radius)
    return _filter(array, rfilter)


if __name__ == '__main__':
    pass



#!/usr/bin/env python

'''Avena is an image processing library and command-line tool.'''


__author__ = 'Mansour Moufid'
__copyright__ = 'Copyright 2014, 2015, Mansour Moufid'
__license__ = 'ISC'
__version__ = '0.5'
__email__ = 'mansourmoufid@gmail.com'
__status__ = 'Development'


if __name__ == '__main__':
    pass



#!/usr/bin/env python

from distutils.core import setup

from hipshot import hipshot


_classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: End Users/Desktop',
    'License :: OSI Approved :: ISC License (ISCL)',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Multimedia :: Graphics',
    'Topic :: Multimedia :: Video',
]

with open('README.rst', 'r') as file:
    _long_description = file.read()

_setup_args = {
    'author':           hipshot.__author__,
    'author_email':     hipshot.__email__,
    'classifiers':      _classifiers,
    'description':      hipshot.__doc__,
    'license':          hipshot.__license__,
    'long_description': _long_description,
    'name':             'Hipshot',
    'url':              'https://bitbucket.org/eliteraspberries/hipshot',
    'version':          hipshot.__version__,
}


if __name__ == '__main__':

    setup(packages=['avena', 'hipshot'], scripts=['scripts/hipshot'],
          **_setup_args)



#!/usr/bin/env python

'''Read and write image files as NumPy arrays'''


from numpy import asarray, float32
from PIL import Image

from . import np
from . import utils


_DEFAULT_DTYPE = float32

_PIL_RGB = {
    'R': 0,
    'G': 1,
    'B': 2,
}


def get_channels(img):
    '''Return a list of channels of an image array.'''
    if utils.depth(img) == 1:
        yield img
    else:
        for i in range(utils.depth(img)):
            yield img[:, :, i]


def read(filename, dtype=_DEFAULT_DTYPE, normalize=True):
    '''Read an image file as an array.'''
    img = Image.open(filename)
    arr = asarray(img, dtype=dtype)
    utils.swap_rgb(arr, _PIL_RGB, to=utils._PREFERRED_RGB)
    if normalize:
        np.normalize(arr)
    return arr


def _pil_save(img, filename):
    pil_img = Image.fromarray(img)
    pil_img.save(filename)
    return


def save(img, filename, random=False, ext=None, normalize=False):
    '''Save an image array and return its path.'''
    if random:
        newfile = utils.rand_filename(filename, ext=ext)
    else:
        newfile = filename
    utils.swap_rgb(img, utils._PREFERRED_RGB, to=_PIL_RGB)
    if normalize:
        np.normalize(img)
    np.clip(img, np._dtype_bounds[str(img.dtype)])
    uint8img = np.to_uint8(img)
    _pil_save(uint8img, newfile)
    return newfile


if __name__ == '__main__':
    pass



#!/usr/bin/env python

from os.path import exists, splitext
from random import randint


_depth = lambda x, y, z=1: z

_invert_dict = lambda d: dict((v, k) for k, v in list(d.items()))

_PREFERRED_RGB = {
    'R': 0,
    'G': 1,
    'B': 2,
}


def depth(array):
    '''Return the depth (the third dimension) of an array.'''
    return _depth(*array.shape)


def rand_filename(filename, ext=None):
    '''Return a unique file name based on the given file name.'''
    file_name, file_ext = splitext(filename)
    if ext is None:
        ext = file_ext
    while True:
        rand_file_name = file_name
        rand_file_name += '-'
        rand_file_name += str(randint(0, 10000))
        rand_file_name += ext
        if not exists(rand_file_name):
            break
    return rand_file_name


def swap_rgb(img, current, to):
    '''Swap the RBG channels of an image array.'''
    if depth(img) == 3 and not current == to:
        current_indices = list(map(current.get, ('R', 'G', 'B')))
        to_indices = list(map(to.get, ('R', 'G', 'B')))
        img[:, :, current_indices] = img[:, :, to_indices]


if __name__ == '__main__':
    pass
